import { Directive, DefineComponent, VNodeProps, AllowedComponentProps, ComponentCustomProps, EmitsOptions, ObjectEmitsOptions, defineComponent } from '@vue/runtime-core';

interface DirectiveUsage<A, E, M extends string> {
  arg?: A
  exp?: E
  modifiers?: M[]
}

type VModelInput<E, EE = E> = DirectiveUsage<
  never,
  E,
  'lazy' | 'trim' | 'number'
> & { _exp?: EE }

declare function resolveDirective<T extends {}, A, B>(
  directives: T,
  name: A,
  cameelName?: B,
): A extends keyof T ? T[A] : B extends keyof T ? T[B] : A

/* eslint-disable @typescript-eslint/prefer-function-type */
type TupleToUnion<T> = T extends Readonly<[infer F, ...infer R]>
  ? F | TupleToUnion<R>
  : T extends Readonly<Array<infer R>>
  ? R
  : never

type UnionToIntersection<U> =
  (U extends any ? (k: U) => void : never) extends ((k: infer I) => void) ? I : never;

type _HalfEquals<A, B> = (
  A extends unknown
  ? (
    B extends unknown
    ? A extends B
    ? B extends A
    ? keyof A extends keyof B
    ? keyof B extends keyof A
    ? A extends object
    ? _DeepHalfEquals<A, B, keyof A> extends true
    ? 1
    : never
    : 1
    : never
    : never
    : never
    : never
    : unknown
  ) extends never
  ? 0
  : never
  : unknown
) extends never
  ? true
  : false
type _DeepHalfEquals<A, B extends A, K extends keyof A> = (
  K extends unknown ? (Equals<A[K], B[K]> extends true ? never : 0) : unknown
) extends never
  ? true
  : false

type Equals<A, B> = _HalfEquals<A, B> extends true
  ? _HalfEquals<B, A>
  : false

type Exact<A, B, Y = unknown, N = never> = Equals<A, B> extends true
  ? Y
  : N

type OnlyEventNames<K extends string | number | symbol> = K extends `on${Capitalize<string>}` ? K : never

// prettier-ignore
type EventName<T extends string> = `on${Capitalize<T>}`

// prettier-ignore
type RemoveOnPrefix<T extends string> = T extends `on${infer K}` ? Uncapitalize<K> : T

type Overloads<T> =
  T extends {
    (...args: infer A1): infer R1;
    (...args: infer A2): infer R2;
    (...args: infer A3): infer R3;
    (...args: infer A4): infer R4;
    (...args: infer A5): infer R5;
    (...args: infer A6): infer R6;
  } ? [
    (...args: A1) => R1,
    (...args: A2) => R2,
    (...args: A3) => R3,
    (...args: A4) => R4,
    (...args: A5) => R5,
    (...args: A6) => R6
  ] : T extends {
    (...args: infer A1): infer R1;
    (...args: infer A2): infer R2;
    (...args: infer A3): infer R3;
    (...args: infer A4): infer R4;
    (...args: infer A5): infer R5;
  } ? [
    (...args: A1) => R1,
    (...args: A2) => R2,
    (...args: A3) => R3,
    (...args: A4) => R4,
    (...args: A5) => R5
  ] : T extends {
    (...args: infer A1): infer R1;
    (...args: infer A2): infer R2;
    (...args: infer A3): infer R3;
    (...args: infer A4): infer R4
  } ? [
    (...args: A1) => R1,
    (...args: A2) => R2,
    (...args: A3) => R3,
    (...args: A4) => R4
  ] : T extends {
    (...args: infer A1): infer R1;
    (...args: infer A2): infer R2;
    (...args: infer A3): infer R3
  } ? [
    (...args: A1) => R1,
    (...args: A2) => R2,
    (...args: A3) => R3
  ] : T extends {
    (...args: infer A1): infer R1;
    (...args: infer A2): infer R2
  } ? [
    (...args: A1) => R1,
    (...args: A2) => R2
  ] : T extends {
    (...args: infer A1): infer R1
  } ? [
    (...args: A1) => R1
  ] : any


type KnownKeys<T> = {
  [K in keyof T as string extends K
  ? never
  : number extends K
  ? never
  : K]: T[K];
}

/* eslint-disable @typescript-eslint/consistent-type-definitions */


type GenericInputModelCheck<B = string> = <E extends B>(
  directives: Array<VModelInput<E>>,
  options?: any,
  checkboxes?: any,
) => any

interface CheckModelDirectiveForDOM {
  radio<E extends TupleToUnion<O>, O>(
    directives: Array<VModelInput<E>>,
    options: O,
    checkboxes: any,
  ): any
  select<E extends EE, O, EE = TupleToUnion<O> & Exact<E, TupleToUnion<O>>>(
    directives: Array<
      VModelInput<E & Exact<E, TupleToUnion<O>>, TupleToUnion<O>>
    >,
    options: O,
    checkboxes: any,
  ): any
  checkbox<E extends Y | N, Y = true, N = false>(
    directives: Array<VModelInput<E>>,
    options: any,
    checkboxes: Readonly<[Y, N]>,
  ): any

  button: GenericInputModelCheck<never>
  color: GenericInputModelCheck
  date: GenericInputModelCheck<string | Date>
  /** @deprecated */
  datetime: GenericInputModelCheck<string | Date>
  'datetime-local': GenericInputModelCheck<string | Date>
  email: GenericInputModelCheck
  file: GenericInputModelCheck<any>
  hidden: GenericInputModelCheck
  image: GenericInputModelCheck<any>
  month: GenericInputModelCheck
  number: GenericInputModelCheck<string | number>
  password: GenericInputModelCheck
  range: GenericInputModelCheck<string | number>
  reset: GenericInputModelCheck<never>
  search: GenericInputModelCheck
  submit: GenericInputModelCheck<never>
  tel: GenericInputModelCheck
  time: GenericInputModelCheck
  text: GenericInputModelCheck
  url: GenericInputModelCheck
  week: GenericInputModelCheck
}

declare const checkModelDirectiveForDOM: CheckModelDirectiveForDOM

interface CheckBuiltinDirective {
  cloak(tag: any, directives: Array<DirectiveUsage<never, never, never>>): any
  html(tag: any, directives: Array<DirectiveUsage<never, string, never>>): any
  memo(tag: any, directives: Array<DirectiveUsage<never, any[], never>>): any
  once(tag: any, directives: Array<DirectiveUsage<never, never, never>>): any
  pre(tag: any, directives: Array<DirectiveUsage<never, never, never>>): any
  show(tag: any, directives: Array<DirectiveUsage<never, boolean, never>>): any
  text(tag: any, directives: Array<DirectiveUsage<never, string, never>>): any
}

declare const checkBuiltinDirective: CheckBuiltinDirective

type MouseEventNames =
  | 'auxclick'
  | 'click'
  | 'contextmenu'
  | 'dbclick'
  | 'mousedown'
  | 'mouseenter'
  | 'mousemove'
  | 'mouseleave'
  | 'mouseout'
  | 'mouseover'
  | 'mouseup'
  | 'wheel'
type PointerEventNames =
  | 'pointerdown'
  | 'pointermove'
  | 'pointerup'
  | 'pointercancel'
  | 'pointerenter'
  | 'pointerleave'
  | 'pointerover'
  | 'pointerout'
type KeyboardEventNames = 'keydown' | 'keypress' | 'keyup'

declare function checkOnDirective<T>(
  tag: T,
): T extends keyof KnownKeys<JSX.IntrinsicElements>
  ? <
      A extends RemoveOnPrefix<OnlyEventNames<keyof JSX.IntrinsicElements[T]>>,
      M extends
        | 'once'
        | 'stop'
        | 'prevent'
        | 'capture'
        | 'self'
        | 'passive'
        | (A extends KeyboardEventNames
            ?
                | 'enter'
                | 'tab'
                | 'delete'
                | 'esc'
                | 'space'
                | 'left'
                | 'right'
                | 'up'
                | 'down'
                | 'ctrl'
                | 'alt'
                | 'shift'
                | 'meta'
                | 'exact'
            : never)
        | (A extends PointerEventNames
            ?
                | 'left'
                | 'right'
                | 'middle'
                | 'ctrl'
                | 'alt'
                | 'shift'
                | 'meta'
                | 'exact'
            : never)
        | (A extends MouseEventNames
            ?
                | 'left'
                | 'right'
                | 'middle'
                | 'ctrl'
                | 'alt'
                | 'shift'
                | 'meta'
                | 'exact'
            : never),
      E extends Get<JSX.IntrinsicElements[T], EventName<A>>
    >(
      directives: Array<DirectiveUsage<A, E, M>>,
    ) => E
  : <
      A extends RemoveOnPrefix<OnlyEventNames<keyof PropsOf<T>>>,
      M extends 'once',
      E extends Get<PropsOf<T>, EventName<A>>
    >(
      directives: Array<DirectiveUsage<A, E, M>>,
    ) => E

declare function checkModelDirective<
  T,
  A extends keyof PropsOf<T>,
  E extends Get<PropsOf<T>, A>
>(tag: T, directives: Array<DirectiveUsage<A, E, never>>): E

declare function checkDirective<
  T,
  A,
  M extends string,
  E,
  D extends Directive<T, E>
>(dir: D, tag: T, directives: Array<DirectiveUsage<A, E, M>>): any
type PropsOf<T> = T extends DefineComponent<
  unknown, // PropsOrPropOptions
  any, // RawBindings
  any, // D
  any, // C
  any, // M
  any, // Mixin
  any, // Extends
  any, // E
  any, // EE
  unknown, // PP
  any, // Props
  any // Defaults
>
  ? InstanceType<T>['$props']
  : {}
type Get<T, K, F = never> = K extends keyof T ? T[K] : F

declare function checkInterpolation(
  value: string | boolean | number | null | undefined,
): string

/**
 * @deprecated
 */
declare function checkInterpolation(value: object | symbol | unknown[]): string

declare function checkSlots<T>(tag: T, slots: any): any

/* eslint-disable @typescript-eslint/no-empty-interface */


declare global {
  namespace JSX {
    interface IntrinsicElements { }
    interface IntrinsicAttributes { }
  }
}

declare function resolveComponent<T extends {}, A, B>(
  components: T,
  name: A,
  pascalName?: B,
): A extends keyof T
  ? T[A]
  : B extends keyof T
  ? T[B]
  : A extends keyof KnownKeys<JSX.IntrinsicElements>
  ? JSX.IntrinsicElements[A] 
  : B extends keyof KnownKeys<JSX.IntrinsicElements>
  ? JSX.IntrinsicElements[B]
  : never

type EmitsToProps<T extends EmitsOptions> = T extends string[] ? {
  [K in string & EventName<T[number]>]?: (...args: any[]) => any;
} : T extends ObjectEmitsOptions ? {
  [K in string & EventName<Capitalize<string & keyof T>>]?: K extends `on${infer C}` ? T[Uncapitalize<C>] extends null ? (...args: any[]) => any : (...args: T[Uncapitalize<C>] extends (...args: infer P) => any ? P : never) => any : never;
} : {};


type TypeEmitsToOptions<T, TU = TupleToUnion<Overloads<T>>> = UnionToIntersection<TU extends ((eventName: infer P, ...rest: infer A) => infer R) ? P extends string ? { [K in P]: (...args: A) => R } : {} : {}>
type ToEmitOptions<T> = T extends (...args: any) => any ? TypeEmitsToOptions<T> & {} : T extends EmitsOptions ? T : {}

declare function defineSetupComponent<P, E, B>(
  props: P,
  emits: E,
  bindings: B,
  options: any,
): DefineComponent<
  {}, // PropsOrPropOptions
  B, // RawBindings
  {}, // D
  {}, // C
  {}, // M
  {}, // Mixin
  {}, // Extends
  {}, // E
  string, // EE
  VNodeProps & AllowedComponentProps & ComponentCustomProps, // PP
  P & EmitsToProps<ToEmitOptions<E>>, // Props
  {} // Defaults
>

declare function renderList<R>(
  source: string,
  handler: (value: string, index: number) => R,
): R[]
declare function renderList<R>(
  source: number,
  handler: (value: number, index: number) => R,
): R[]
declare function renderList<R>(
  source: number,
  handler: (value: number, index: number) => R,
): R[]
declare function renderList<T, R>(
  source: T[],
  handler: (value: T, index: number) => R,
): R[]
declare function renderList<T extends Readonly<unknown[]>, R>(
  source: T,
  handler: (value: TupleToUnion<T>, index: number) => R,
): R[]
declare function renderList<T, R>(
  source: Iterable<T>,
  handler: (value: T, index: number) => R,
): R
declare function renderList<T extends object, R>(
  source: T,
  handler: <K extends keyof T>(value: T[K], key: K, index: number) => R,
): R[]

// -- renderSlot --
declare function renderSlot<
  S extends Record<string, (...args: any) => any>,
  K extends keyof S
>(source: S, name: K, props: Parameters<S[K]>[0]): any

/* eslint-disable @typescript-eslint/no-empty-interface */


type version = '3.x'

declare global {
  namespace VueDX {
    namespace internal {
      export { resolveComponent, resolveDirective }
      export { renderList, renderSlot }
      export { defineComponent, defineSetupComponent }
      export {
        checkInterpolation,
        checkSlots,
        checkDirective,
        checkOnDirective,
        checkModelDirective,
        checkModelDirectiveForDOM,
        checkBuiltinDirective,
      }
    }
  }
}

export { version };
